﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace ASW
{
    public partial class frmMain : Form
    {
        public SerialPort Conn = null;
        public char[] serialBuffer;
        public int iSB = 0;
        public bool ClientCycling = false;
        public char[] MSG_IDENT;                //  identity
        public char[] MSG_ANNOUNCE;             //  broadcast existence
        public char[] MSG_READY;                //  ready state
        public char[] MSG_BUSY;                 //  busy state
        public char[] MSG_SETNAME;              //  set cycle name
        public char[] MSG_SETDESCRIPTION;       //  set cycle description
        public char[] MSG_SETPOWER;             //  set cycle power
        public char[] MSG_SETTIME;              //  set cycle time
        public char[] MSG_DEBUG;              //  set cycle time
        public char MSG_BREAK;                //  message break
        public char MSG_END;                  //  end of message
        public bool Saved = false;
        private bool focusBreak = false;
        static bool m_SetThreadedInfo = false;
        static bool m_SetThreadedInfoPBOnly = false;
        static string m_ConState = "";
        static string m_StatusText = "";
        static Color m_StatusColor = Color.Black;
        static int m_StatusTimeout = 0;
        static Color m_ConColor = Color.DarkRed;
        static int m_PBMin = 0;
        static int m_PBMax = 7;
        static int m_PBVal = 0;
        static bool m_PBVis = false;
        static bool m_SendEnabled = false;
        static string m_CycleName = "";
        static string m_CycleDescription = "";
        static string m_CyclePower = "1";
        static string m_CycleTime = "1000";

        public frmMain()
        {
            InitializeComponent();
        }

        private void frmMain_Load(object sender, EventArgs e)
        {
            this.serialBuffer = new char[30];
            this.MSG_IDENT = new char[] {'A','S','W'};
            this.MSG_ANNOUNCE = new char[] {'A','N','N','O','U','N','C','E'};
            this.MSG_READY = new char[] {'R','E','A','D','Y'};
            this.MSG_BUSY = new char[] {'B','U','S','Y'};
            this.MSG_SETNAME = new char[] {'S','E','T','N'};
            this.MSG_SETDESCRIPTION = new char[] {'S','E','T','D'};
            this.MSG_SETPOWER = new char[] { 'S', 'E', 'T', 'P' };
            this.MSG_SETTIME = new char[] { 'S', 'E', 'T', 'T' };
            this.MSG_DEBUG = new char[] { 'D', 'E', 'B', 'U', 'G' };
            this.MSG_BREAK = ':';
            this.MSG_END = ';';
            this.UpdatePortsList();
            LoadDB();
        }

        void frmMain_FormClosing(object sender, System.Windows.Forms.FormClosingEventArgs e)
        {
            if (this.Saved == false)
            {
                switch (MessageBox.Show("The database has been changed if you do not save it your changes will be lost.  Would you like to save the database before closing?", "Database Modified", MessageBoxButtons.YesNoCancel))
                {
                    case DialogResult.Yes:
                        this.SaveDB();
                        break;
                    case DialogResult.No:
                        break;
                    case DialogResult.Cancel:
                        e.Cancel = true;
                        break;
                }
            }
        }

        private void LoadDB()
        {
            if (System.IO.File.Exists(Application.StartupPath + "\\db.dat"))
            {
                string[] lines = System.IO.File.ReadAllLines(Application.StartupPath + "\\db.dat");
                string[] tA;
                foreach (string line in lines)
                {
                    tA = line.Split('\t');
                    if (tA.Length > 3)
                    {
                        lvData.Items.Add(tA[0]);
                        lvData.Items[lvData.Items.Count - 1].SubItems.Add(tA[1]);
                        lvData.Items[lvData.Items.Count - 1].SubItems.Add(tA[2]);
                        lvData.Items[lvData.Items.Count - 1].SubItems.Add(tA[3]);
                    }
                }
                if (lvData.Items.Count > 0)
                {
                    lvData.Items[0].Selected = true;
                }
                else
                {
                    txtCycleName.Text = "TEMP CYCLE";
                    txtCycleDescription.Text = "REPLACE ME";
                    cbCyclePower.SelectedIndex = 0;
                    txtCycleTime.Text = "1000";
                }
            }
            else
            {
                txtCycleName.Text = "TEMP CYCLE";
                txtCycleDescription.Text = "REPLACE ME";
                cbCyclePower.SelectedIndex = 0;
                txtCycleTime.Text = "1000";
            }
            this.Saved = true;
        }
        private void SaveDB()
        {
            StringBuilder sB = new StringBuilder();
            foreach (ListViewItem lvI in this.lvData.Items)
            {
                sB.AppendLine(lvI.Text + '\t'.ToString() + lvI.SubItems[1].Text + '\t'.ToString() + lvI.SubItems[2].Text + '\t'.ToString() + lvI.SubItems[3].Text);
            }
            System.IO.File.WriteAllText(Application.StartupPath + "\\db.dat", sB.ToString());
            this.Saved = true;
        }

        private void UpdatePortsList()
        {
            string[] ports = SerialPort.GetPortNames();
            foreach (string port in ports)
            {
                if (this.cbPort.Items.Contains(port) == false)
                {
                    this.cbPort.Items.Add(port);
                }
            }
            if (this.cbPort.Items.Count > 0)
            {
                this.cbPort.SelectedIndex = 0;
            }
        }

        public void TransmitCycle()
        {
            char[] cB = new char[1] { this.MSG_BREAK };
            char[] cE = new char[1] { this.MSG_END };
            char[] cN = m_CycleName.ToCharArray();
            char[] cD = m_CycleDescription.ToCharArray();
            char[] cP = m_CyclePower.ToCharArray();
            char[] cT = m_CycleTime.ToCharArray();
            Conn.Write(this.MSG_BUSY, 0, this.MSG_BUSY.Length);
            Conn.Write(cE, 0, cE.Length);
            Conn.Write(this.MSG_SETNAME, 0, this.MSG_SETNAME.Length);
            Conn.Write(cB, 0, cB.Length);
            Conn.Write(cN, 0, cN.Length);
            Conn.Write(cE, 0, cE.Length);
            Conn.Write(this.MSG_SETDESCRIPTION, 0, this.MSG_SETDESCRIPTION.Length);
            Conn.Write(cB, 0, cB.Length);
            Conn.Write(cD, 0, cD.Length);
            Conn.Write(cE, 0, cE.Length);
            Conn.Write(this.MSG_SETPOWER, 0, this.MSG_SETPOWER.Length);
            Conn.Write(cB, 0, cB.Length);
            Conn.Write(cP, 0, cP.Length);
            Conn.Write(cE, 0, cE.Length);
            Conn.Write(this.MSG_SETTIME, 0, this.MSG_SETTIME.Length);
            Conn.Write(cB, 0, cB.Length);
            Conn.Write(cT, 0, cT.Length);
            Conn.Write(cE, 0, cE.Length);
            Conn.Write(this.MSG_READY, 0, this.MSG_READY.Length);
            Conn.Write(cE, 0, cE.Length);
            m_StatusText = "Cycle Transmitted";
            m_StatusColor = Color.Black;
            m_StatusTimeout = 2500;
            m_SetThreadedInfo = true;
        }

        private void cmdConnectDisconnect_Click(object sender, EventArgs e)
        {
            if (this.cmdConnectDisconnect.Text == "&Connect")
            {
                try
                {
                    this.cmdSendToASW.Enabled = false;
                    this.Conn = new SerialPort(cbPort.SelectedItem.ToString(), 9600, Parity.None, 8, StopBits.One);
                    this.Conn.DataReceived += new SerialDataReceivedEventHandler(Conn_DataReceived);
                    this.Conn.Disposed += new EventHandler(Conn_Disposed);
                    this.Conn.Open();
                    tmrPortRefresh.Enabled = false;
                    this.cmdConnectDisconnect.Text = "Dis&connect";
                    this.tsslConnectionState.Text = "Listening";
                    this.tsslStatus.Text = "Attempting to connect.";
                    this.tsslConnectionState.ForeColor = Color.SteelBlue;
                    this.tspbProgress.Visible = false;
                }
                catch (Exception ex)
                {
                    tsslStatus.Text = "Unable to connect to the selected serial port.";
                    tsslStatus.ForeColor = Color.Red;
                    ClearStatus(3500);
                }
            }
            else
            {
                this.Conn.Close();
                this.Conn.Dispose();
                this.Conn = null;
                this.cmdSendToASW.Enabled = false;
                tmrPortRefresh.Enabled = true;
                this.cmdConnectDisconnect.Text = "&Connect";
                this.tsslConnectionState.Text = "Not Connected";
                this.tsslConnectionState.ForeColor = Color.DarkRed;
                this.tsslStatus.Text = "";
                this.tspbProgress.Visible = false;
            }
        }

        void Conn_Disposed(object sender, EventArgs e)
        {
            this.cmdSendToASW.Enabled = false;
            tmrPortRefresh.Enabled = true;
            this.cmdConnectDisconnect.Text = "&Connect";
            this.tsslConnectionState.Text = "Not Connected";
            this.tsslConnectionState.ForeColor = Color.DarkRed;
            this.tsslStatus.Text = "";
            this.tspbProgress.Visible = false;
        }

        public void ClearStatus(int timeout)
        {
            this.tmrProgressClear.Enabled = false;
            this.tmrProgressClear.Interval = timeout;
            this.tmrProgressClear.Enabled = true;
        }

        /*  
         *  This function is a hack to facilitate ease of porting from an arduino environment
         *  I in no way recommend using it in anything else, its not even close to the real strncmp.
         */
        private bool strncmp(char[] str1, int start, char[] str2, int length)
        {
            if (((str1.Length - start) >= length) && (str2.Length >= length))
            {
                for (int i = 0; i < length; i++)
                {
                    if (str1[start + i] != str2[i]) { return (false); }
                }
                return (true);
            }
            else { return (false); }
        }

        private int GetMessage(char[] data, int start, int length)
        {
            if (length > 2)
            {
                if (strncmp(data, start, MSG_IDENT, 3)) { return (1); }
                else if (length > 3)
                {
                    if (strncmp(data, start, MSG_BUSY, 4)) { return (4); }
                    else if (strncmp(data, start, MSG_SETNAME, 4)) { return (5); }
                    else if (strncmp(data, start, MSG_SETDESCRIPTION, 4)) { return (6); }
                    else if (strncmp(data, start, MSG_SETPOWER, 4)) { return (7); }
                    else if (strncmp(data, start, MSG_SETTIME, 4)) { return (8); }
                    else if (length > 4)
                    {
                        if (strncmp(data, start, MSG_READY, 5)) { return (3); }
                        if (strncmp(data, start, MSG_DEBUG, 5)) { return (9); }
                        else if (length > 7)
                        {
                            if (strncmp(data, start, MSG_ANNOUNCE, 8)) { return (2); }
                        }
                    }
                }
            }
            return (0);
        }
        
        private void DataArrival(char[] data, int length, int break0, int break1, int break2, int break3) {
            int iMsg = 0;
            if(break0 > 0) {
                iMsg = GetMessage(data, 0, break0);
            } else {
                iMsg = GetMessage(data, 0, length);
            }
            char[] cE = new char[1] { this.MSG_END };
            switch(iMsg) {
                case 1:      //  Identity
                    //  unused
                    break;
                case 2:      //  Announce
                    this.Conn.Write(this.MSG_IDENT, 0, this.MSG_IDENT.Length);
                    this.Conn.Write(cE, 0, cE.Length);
                    m_ConState = "Connected";
                    m_StatusText = "";
                    m_StatusColor = Color.Black;
                    m_ConColor = Color.Blue;
                    m_PBVis = false;
                    m_SendEnabled = true;
                    m_SetThreadedInfo = true;
                    this.TransmitCycle();
                    break;
                case 3:      //  Ready
                    this.ClientCycling = false;
                    m_StatusText = "Cycle Complete";
                    m_StatusColor = Color.Black;
                    m_StatusTimeout = 3500;
                    m_SetThreadedInfo = true;
                    break;
                case 4:      //  Busy
                    m_StatusText = "Cycle Started";
                    m_StatusColor = Color.Red;
                    m_SetThreadedInfo = true;
                    break;
                case 5:      //  Set Name
                    //  unused
                    break;
                case 6:      //  Set Description
                    //  unused
                    break;
                case 7:      //  Set Power
                    //  unused
                    break;
                case 8:      //  Set Time
                    //  unused
                    break;
                case 9:      //  Debug Message
                    Console.WriteLine(data, 0, length);
                    break;
            }
        }

        private void processSerialBuffer() {
            int[] b = new int[4];
            b[0] = -1; b[1] = -1; b[2] = -1; b[3] = -1;
            int iB = 0;
            int i = 0;
            while((i < iSB) && (i < 30)) {
                if(serialBuffer[i] == MSG_END) {
                    break;
                } else if(serialBuffer[i] == MSG_BREAK) {
                    b[iB] = i;
                    iB++;
                }
                i++;
            }
            if(i < 30) {
                if(i > 0) {
                    char[] tB = new char[i];
                    for(int j = 0; j < i; j++) {
                        tB[j] = serialBuffer[j];
                    }
                    DataArrival(tB, i, b[0], b[1], b[2], b[3]);
                    if(i < 29) {
                        for(int j = 0; j < (29 - i); j++) {
                            serialBuffer[j] = serialBuffer[j + (i + 1)];
                        }
                        iSB = iSB - (i + 1);
                    } else {
                        for(int i2 = 0; i2 < 30; i2++) { serialBuffer[i2] = ' '; }
                        iSB = 0;
                    }
                } else {
                    for(int j = 0; j < 29; j++) {
                        serialBuffer[j] = serialBuffer[j + 1];
                    }
                    serialBuffer[29] = ' ';
                    iSB--;
                }
            } else {
                if(iSB > 29) {
                    for(int i2 = 0; i2 < 30; i2++) { serialBuffer[i2] = ' '; }
                    iSB = 0;
                }
            }
        }

        private void Conn_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            byte iB;
            while (Conn.BytesToRead > 0)
            {
                iB = Convert.ToByte(Conn.ReadByte());
                if (iB != 0)
                {
                    if (iSB < 30)
                    {
                        serialBuffer[iSB] = Convert.ToChar(iB);
                        iSB++;
                        if (Convert.ToChar(iB) == ';') { processSerialBuffer(); }
                    }
                    else
                    {
                        processSerialBuffer();
                        serialBuffer[iSB] = Convert.ToChar(iB);
                        iSB++;
                    }
                }
            }
        }

        private void tmrProgressClear_Tick(object sender, EventArgs e)
        {
            tmrProgressClear.Enabled = false;
            this.tsslStatus.Text = "";
            this.tsslStatus.ForeColor = Color.Black;
        }

        private void lvData_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (lvData.SelectedItems.Count > 0)
            {
                txtCycleName.Text = lvData.SelectedItems[0].Text;
                txtCycleDescription.Text = lvData.SelectedItems[0].SubItems[1].Text;
                cbCyclePower.SelectedIndex = Convert.ToInt32(lvData.SelectedItems[0].SubItems[2].Text) - 1;
                txtCycleTime.Text = lvData.SelectedItems[0].SubItems[3].Text;
            }
        }

        private void cmdDBRemove_Click(object sender, EventArgs e)
        {
            if (lvData.SelectedItems.Count > 0)
            {
                lvData.Items.Remove(lvData.SelectedItems[0]);
                if (lvData.SelectedItems.Count < 1)
                {
                    if (lvData.Items.Count > 0)
                    {
                        lvData.Items[0].Selected = true;
                    }
                    else
                    {
                        txtCycleName.Text = "TEMP CYCLE";
                        txtCycleDescription.Text = "REPLACE ME";
                        cbCyclePower.SelectedIndex = 0;
                        txtCycleTime.Text = "1000";
                    }
                }
            }
            this.Saved = false;
        }

        private void cmdClearFields_Click(object sender, EventArgs e)
        {
            txtCycleName.Text = "";
            txtCycleDescription.Text = "";
            cbCyclePower.SelectedIndex = 0;
            txtCycleTime.Text = "";
        }

        private void cmdDBAdd_Click(object sender, EventArgs e)
        {
            lvData.Items.Add(txtCycleName.Text);
            lvData.Items[lvData.Items.Count - 1].SubItems.Add(txtCycleDescription.Text);
            lvData.Items[lvData.Items.Count - 1].SubItems.Add(Convert.ToString(cbCyclePower.SelectedIndex + 1));
            lvData.Items[lvData.Items.Count - 1].SubItems.Add(txtCycleTime.Text);
            this.Saved = false;
        }

        private void cmdDBLoad_Click(object sender, EventArgs e)
        {
            this.LoadDB();
        }

        private void cmdDBSave_Click(object sender, EventArgs e)
        {
            this.SaveDB();
        }

        private void tmrPortRefresh_Tick(object sender, EventArgs e)
        {
            this.UpdatePortsList();
        }

        private void cmdSendToASW_Click(object sender, EventArgs e)
        {
            TransmitCycle();
        }

        void txtCycleTime_LostFocus(object sender, System.EventArgs e)
        {
            try
            {
                long l = Convert.ToInt64(txtCycleTime.Text);
            }
            catch (Exception ex)
            {
                this.focusBreak = true;
                MessageBox.Show("The cycle time entered is invalid.  There are 1,000 milliseconds in a second.  Cycle time must be a numeric value.", "Invalid Cycle Time");
                txtCycleTime.Focus();
                txtCycleTime.SelectAll();
                this.focusBreak = false;
            }
        }

        private void tmrThreadHappy_Tick(object sender, EventArgs e)
        {
            if (m_SetThreadedInfo == true)
            {
                m_SetThreadedInfo = false;
                if (this.tsslConnectionState.Text != m_ConState) { this.tsslConnectionState.Text = m_ConState; }
                if (this.tsslStatus.Text != m_StatusText) { this.tsslStatus.Text = m_StatusText; }
                if (this.tsslConnectionState.ForeColor != m_ConColor) { this.tsslConnectionState.ForeColor = m_ConColor; }
                if (this.tspbProgress.Visible != m_PBVis) { this.tspbProgress.Visible = m_PBVis; }
                if (this.cmdSendToASW.Enabled != m_SendEnabled) { this.cmdSendToASW.Enabled = m_SendEnabled; }
                if (this.tsslStatus.ForeColor != m_StatusColor) { this.tsslStatus.ForeColor = m_StatusColor; }
                if (m_StatusTimeout > 0) { this.ClearStatus(m_StatusTimeout); }
            }
            if (m_SetThreadedInfoPBOnly == true)
            {
                m_SetThreadedInfoPBOnly = false;
                if (this.tspbProgress.Minimum != m_PBMin) { this.tspbProgress.Minimum = m_PBMin; }
                if (this.tspbProgress.Maximum != m_PBMax) { this.tspbProgress.Maximum = m_PBMax; }
                if (this.tspbProgress.Value != m_PBVal) { this.tspbProgress.Value = m_PBVal; }
                if (this.tspbProgress.Visible != m_PBVis) { this.tspbProgress.Visible = m_PBVis; }
            }
        }

        private void txtCycleName_TextChanged(object sender, EventArgs e)
        {
            m_CycleName = this.txtCycleName.Text;
        }

        private void txtCycleDescription_TextChanged(object sender, EventArgs e)
        {
            m_CycleDescription = this.txtCycleDescription.Text;
        }

        private void cbCyclePower_SelectedIndexChanged(object sender, EventArgs e)
        {
            m_CyclePower = this.cbCyclePower.Text;
        }

        private void txtCycleTime_TextChanged(object sender, EventArgs e)
        {
            try
            {
                long l = Convert.ToInt64(this.txtCycleTime.Text);
                m_CycleTime = this.txtCycleTime.Text;
            }
            catch (Exception ex)
            {
            }
        }
    }
}
